"""
Abstraction of command-line programs which produce data files.
"""
import abc
from os import PathLike
from typing import Sequence, Union


class AbstractDataCommand(abc.ABC):
    """
    An `AbstractDataCommand` represents a command-line program which can produce
    output data to a given path.

    Subclasses of `AbstractDataCommand` must be
    [hashable](https://docs.python.org/3/library/typing.html#typing.Hashable)
    and immutable, because `AbstractDataCommand` objects are used as dictionary
    keys by `civet.memoization.Memoizer`. It's recommended to use
    [frozen dataclasses](https://docs.python.org/3/library/dataclasses.html#frozen-instances).
    """

    @abc.abstractmethod
    def command(self, output: str | PathLike) -> Sequence[Union[str, PathLike, 'AbstractDataCommand']]:
        ...

    @property
    def preferred_suffix(self) -> str:
        """
        Preferred output path suffix for this command.
        """
        return ''



"""
Wrappers around
[`subprocess.run`](https://docs.python.org/3/library/subprocess.html#subprocess.run).
"""

import subprocess as sp
from os import PathLike
from typing import Sequence, Callable

Shell = Callable[[Sequence[str | PathLike]], None]


def subprocess_run(cmd: Sequence[str | PathLike]) -> None:
    """
    Alias for `subprocess.run(cmd, check=True)`
    """
    sp.run(cmd, check=True)


def quiet(cmd: Sequence[str | PathLike]) -> None:
    """
    Similar to `subprocess_run` but output streams are piped to `/dev/null`.
    """
    sp.run(cmd, check=True, stdout=sp.DEVNULL, stderr=sp.STDOUT)



"""
Base classes for file types.
"""
import abc
from os import PathLike
from typing import Sequence, TypeVar, Generic, Callable
from dataclasses import dataclass
from civet.abstract_data import AbstractDataCommand
from civet.memoization import Session
from civet.shells import Shell, subprocess_run


@dataclass(frozen=True)
class DataSource(AbstractDataCommand, abc.ABC):
    """
    A `DataSource` provides the `DataSource.save` method to `civet.abstract_data.AbstractDataCommand`,
    which is an alias for typical usage of `civet.memoization.Session`.
    """
    def save(self, output: str | PathLike,
             require_output: bool = True,
             shell: Shell = subprocess_run) -> None:
        r"""
        Save the result of this command to the given output path.

        The behavior of `save` can be customized by specifying `shell` as different
        wrappers around `subprocess.run`. For example, to save the logs of subprocesses:

        ```python
        from pathlib import Path
        import subprocess as sp

        log_path = Path('output.log')
        with log_path.open('wb') as log_file:
            def saves_log_shell(cmd):
                log_file.write(b'Running: ')
                log_file.write(str(cmd).encode('utf-8'))
                log_file.write(b'\n')
                log_file.flush()
                sp.run(cmd, stdout=log_file, stderr=sp.STDOUT, check=True)
            GenericSurface('input.obj').slide_left().save('lefter.obj', shell=saves_log_shell)
        ```
        """
        with Session(require_output, shell) as s:
            s.save(self, output)


_D = TypeVar('_D', bound='DataFile')


@dataclass(frozen=True)
class DataFile(DataSource, Generic[_D], abc.ABC):
    """
    A `DataFile` represents a file type. It can wrap input files of that file type,
    or commands which produce results of that file type.
    """

    input: str | PathLike | AbstractDataCommand

    def command(self, output: str | PathLike) -> Sequence[str | PathLike | AbstractDataCommand]:
        return 'cp', self.input, output

    def create_command(self, command: Callable[[str | PathLike], Sequence[str | PathLike | AbstractDataCommand]]
                       ) -> _D:
        """
        Chain a command which produces output of the same type as this `DataFile`.
        """
        class Intermediate(self.__class__):
            def command(self, output: str | PathLike) -> Sequence[str | PathLike | AbstractDataCommand]:
                return command(output)
        return Intermediate(self)



"""
Submodule for more specific types which are relevant to
surface-extraction subroutines of the CIVET pipeline.
"""

from civet.extraction.side import Side
from civet.extraction.hemisphere import HemisphereMask
from civet.extraction.surfaces import RegularSurface, IrregularSurface, Tetra
from civet.extraction.starting_models import SurfaceModel

__all__ = [
    'HemisphereMask',
    'Side',
    'RegularSurface',
    'IrregularSurface',
    'Tetra',
    'SurfaceModel'
]



"""
Defines types related to surface meshes (`.obj` file format).
"""
from civet.xfm import TransformableMixin
from civet.minc import GenericMask
from typing import TypeVar, Generic, ClassVar
from dataclasses import dataclass

_S = TypeVar('_S', bound='GenericSurface')
_M = TypeVar('_M', bound=GenericMask)


@dataclass(frozen=True)
class GenericSurface(TransformableMixin[_S], Generic[_S]):
    preferred_suffix: ClassVar[str] = '.obj'
    transform_program: ClassVar[str] = 'transform_objects'

    def surface_mask2(self, in_volume: _M) -> _M:
        def command(output):
            return 'surface_mask2', in_volume, self, output
        return in_volume.create_command(command)

    def adapt_object_mesh(self, target_points: int, n_iterations: int, n_adapt: int, n_adapt_smooth: int) -> _S:
        def command(output):
            return 'adapt_object_mesh', self, output, \
                str(target_points), str(n_iterations), str(n_adapt), str(n_adapt_smooth)
        return self.create_command(command)


class Surface(GenericSurface['Surface']):
    """
    Represents a polygonal mesh of a brain surface in `.obj` file format.
    """
    pass



"""
Defines types and operations related to MINC files.
"""
from dataclasses import dataclass
from os import PathLike
from typing import Literal, TypeVar, Generic, Optional

from civet.bases import DataFile
from civet.extraction.kernels import ngh_count_kernel
from civet.xfm import TransformProgram, TransformableMixin

_M = TypeVar('_M', bound='GenericMinc')
_V = TypeVar('_V', bound='GenericMinc')


@dataclass(frozen=True)
class GenericMinc(TransformableMixin[_M], DataFile[_M], Generic[_M]):

    preferred_suffix = '.mnc'
    transform_program: TransformProgram = 'transform_volume'

    def mincresample(self, like_volume: _V) -> _V:
        def command(output):
            return (
                'mincresample', '-clobber', '-quiet',
                '-like', like_volume, self, output
            )
        return like_volume.create_command(command)

    def reshape255(self) -> 'GenericMask':
        def command(output):
            return (
                'mincreshape', '-quiet', '-clobber', '-unsigned', '-byte',
                '-image_range', '0', '255', '-valid_range', '0', '255',
                self, output
            )
        return GenericMask(self).create_command(command)

    def resamplef64(self, *extra_flags: str) -> 'GenericFloatMinc':
        def command(output):
            return 'mincresample', '-quiet', '-double', *extra_flags, self, output
        return GenericFloatMinc(self).create_command(command)


class MincVolume(GenericMinc['MincVolume']):
    """
    A `MincVolume` represents a volume (`.mnc`).
    """
    pass


_MA = TypeVar('_MA', bound='GenericMask')


class GenericMask(GenericMinc[_MA], Generic[_MA]):

    def dilate_volume(self, dilation_value: int, neighbors: Literal[6, 26], n_dilations: int) -> _MA:
        def command(output):
            return (
                'dilate_volume', self, output,
                str(dilation_value), str(neighbors), str(n_dilations)
            )
        return self.create_command(command)

    def minccalc_u8(self, expression: str, *other_volumes: 'GenericMask') -> _MA:
        def command(output):
            return (
                'minccalc', '-clobber', '-quiet',
                '-unsigned', '-byte',
                '-expression', expression,
                self, *other_volumes, output
            )
        return self.create_command(command)

    def mincdefrag(self, label: int, stencil: Literal[6, 19, 27], max_connect: Optional[int] = None) -> _MA:
        def command(output):
            cmd = ['mincdefrag', self, output, str(label), str(stencil)]
            if max_connect is not None:
                cmd.append(str(max_connect))
            return cmd
        return self.create_command(command)

    def reshape_bbox(self) -> _MA:
        def command(output):
            return 'mincreshape_bbox_helper', self, output
        return self.create_command(command)

    def mincmorph_convolve_u8(self, kernel: str | PathLike = ngh_count_kernel) -> _MA:
        def command(output):
            return (
                'mincmorph',
                '-unsigned', '-byte',
                '-convolve',
                '-kernel', kernel,
                self, output
            )
        return self.create_command(command)


class Mask(GenericMask['Mask']):
    """
    A `Mask` represents a volume (`.mnc`) with discrete intensities (segmentation volume or brain mask).
    """
    pass


_F = TypeVar('_F', bound='GenericFloatMinc')


class GenericFloatMinc(GenericMinc[_F], Generic[_F]):
    def mincblur(self, fwhm: float) -> _F:
        # result is not a binary mask, it has float values in [0, 1],
        # maybe define a unique type?
        def command(output):
            return 'mincblur_correct_name.sh', '-quiet', '-fwhm', str(fwhm), self, output
        return self.create_command(command)


class FloatMinc(GenericFloatMinc['FloatMinc']):
    pass



"""
Data from the `$MNI_DATAPATH/surface-extraction` directory.
"""
from civet.extraction.surfaces import RegularSurface
from civet.globals import MNI_DATAPATH


class SurfaceModel(RegularSurface['SurfaceModel']):
    """
    Represents a surface data file from the `$MNI_DATAPATH/surface-extraction` directory.
    """
    @classmethod
    def get_model(cls, name: str) -> 'SurfaceModel':
        return cls(MNI_DATAPATH / 'surface-extraction' / name)


WHITE_MODEL_320 = SurfaceModel.get_model('white_model_320.obj')



"""
Types related to xfm transformations
(`param2xfm`, `transform_objects`, `transform_volume`).
"""
import abc
from os import PathLike
from typing import Sequence, TypeVar, Literal
from civet.abstract_data import AbstractDataCommand
from civet.bases import DataSource, DataFile
from enum import Enum
from dataclasses import dataclass


class Transformations(Enum):
    CENTER = '-center'
    TRANSLATION = '-translation'
    ROTATIONS = '-rotations'
    SCALES = '-scales'
    SHEARS = '-shears'


@dataclass(frozen=True)
class Xfm(DataSource):
    """
    Represents a `.xfm` file created by `param2xfm`.

    `Xfm` can only represent a single `param2xfm` option.
    Would be nice to support multiple transformations in go, but it's not implemented.
    """

    transformation: Transformations
    x: float
    y: float
    z: float
    preferred_suffix = '.xfm'

    def command(self, output: str | PathLike) -> Sequence[str | PathLike | AbstractDataCommand]:
        return 'param2xfm', self.transformation.value, str(self.x), str(self.y), str(self.z), output


TransformProgram = Literal['transform_objects', 'transform_volume']
_T = TypeVar('_T', bound='TransformableMixin')


@dataclass(frozen=True)
class TransformableMixin(DataFile[_T], abc.ABC):
    @property
    @abc.abstractmethod
    def transform_program(self) -> TransformProgram:
        """
        Specifies which program is used to transform objects of this type.
        """
        ...

    def append_xfm(self, xfm: Xfm) -> _T:
        def command(output: str | PathLike) -> Sequence[str | PathLike | AbstractDataCommand]:
            return self.transform_program, self, xfm, output
        return self.create_command(command)

    def flip_x(self) -> _T:
        """
        Flip this surface along the *x*-axis.
        """
        return self.append_xfm(Xfm(Transformations.SCALES, -1, 1, 1))

    def translate_x(self, x: float) -> _T:
        """
        Translate this surface along the *x*-axis.
        """
        return self.append_xfm(Xfm(Transformations.TRANSLATION, x, 0, 0))

    def slide_left(self) -> _T:
        """
        Translate this surface 25 units to the left.
        """
        return self.translate_x(-25)

    def slide_right(self) -> _T:
        """
        Translate this surface 25 units to the right.
        """
        return self.translate_x(25)



"""
Classes related to pre- and post-processing for
`sphere_mesh` (the implementation of marching-cubes
in CIVET-2.1.1).
"""

from dataclasses import dataclass
from os import PathLike
from typing import Sequence, Optional

from civet.abstract_data import AbstractDataCommand
from civet.extraction.starting_models import SurfaceModel, WHITE_MODEL_320
from civet.extraction.surfaces import IrregularSurface
from civet.minc import Mask, GenericMask
from civet.extraction.side import Side


class HemisphereMask(GenericMask['HemisphereMask']):
    """
    Represents a binary mask of a brain hemisphere (either left or right).
    """

    def just_sphere_mesh(self, side: Optional[Side] = None, subsample: bool = False) -> IrregularSurface:
        """
        Just run `sphere_mesh`, which produces a mesh with non-standard connectivity.
        """
        model = self.get_model_for(side)
        return self.prepare_for_sphere_mesh(model).sphere_mesh(subsample)

    def smoothen_using_mincmorph(self, iterations: int = 5, lower: float = 2.5, upper: float = 4.5) -> 'HemisphereMask':
        """
        Simplify connectivity of this mask using mincmorph.
        This is a preprocessing step to `sphere_mesh`.

        https://github.com/aces/surface-extraction/blob/7c9c5987a2f8f5fdeb8d3fd15f2f9b636401d9a1/scripts/marching_cubes.pl.in#L166-L184
        """
        if not isinstance(iterations, int) or iterations < 0:
            raise ValueError(f'iterations {iterations} < 0')

        smoother = self

        for _ in range(iterations):
            ngh_count = smoother.mincmorph_convolve_u8()
            expression = f'if(A[1]<{lower})' r'{0}else{' f'if(A[1]>{upper})' r'{1}else{A[0]}}'
            smoother = self.minccalc_u8(expression, ngh_count)

        return smoother

    def prepare_for_sphere_mesh(self, initial_model: SurfaceModel) -> 'SphereMeshMask':
        """
        Prepare this mask for `sphere_mesh`.

        https://github.com/aces/surface-extraction/blob/7c9c5987a2f8f5fdeb8d3fd15f2f9b636401d9a1/scripts/marching_cubes.pl.in#L189-L207

        Parameters
        ----------
        initial_model: SurfaceModel
            bounds
        """
        filled = self.minccalc_u8('out=1')
        surface_mask_vol = initial_model.surface_mask2(filled)
        resampled = surface_mask_vol.mincresample(filled)
        overlap = self.minccalc_u8('if(A[0]>0.5||A[1]>0.5){1}else{0}', resampled)
        dilated = overlap.dilate_volume(1, 26, 1)
        added = self.minccalc_u8('A[0]+A[1]', dilated)
        defragged = added.reshape255().mincdefrag(2, 19)
        return SphereMeshMask(defragged)

    @staticmethod
    def get_model_for(side: Optional[Side] = None) -> SurfaceModel:
        """
        Transform `WHITE_MODEL_320` as necessary in preparation for use with `sphere_mesh`.

        https://github.com/aces/surface-extraction/blob/7c9c5987a2f8f5fdeb8d3fd15f2f9b636401d9a1/scripts/marching_cubes.pl.in#L118-L135
        """
        initial_model = WHITE_MODEL_320
        if side is Side.LEFT:
            return initial_model.slide_left()
        elif side is Side.RIGHT:
            return initial_model.flip_x().slide_right()
        else:
            return initial_model


@dataclass(frozen=True)
class SphereMeshMask:
    """
    Represents a mask which is suitable input to the `sphere_mesh` program.
    """

    sphere_mesh_mask: Mask
    """
    `sphere_mesh` expects a `.mnc` where the actual brain hemisphere mask
    is labeled with `2`. The `2` region should be contained within an
    oval region labeled `1`, which is produced by running `surface_mask2`
    on a `SurfaceModel`.
    """

    def sphere_mesh(self, subsample: bool = False) -> IrregularSurface:
        class SphereMeshSurface(IrregularSurface):
            def command(self, output: str | PathLike
                        ) -> Sequence[str | PathLike | AbstractDataCommand]:
                if subsample:
                    return 'sphere_mesh', self.input, output, '-subsample'
                return 'sphere_mesh', self.input, output
        return SphereMeshSurface(self.sphere_mesh_mask)
